#include "simdutf/common_defs.h"

namespace simdutf {
namespace tests {
namespace reference {

simdutf_warn_unused bool validate_utf8(const char *buf, size_t len) noexcept;

} // namespace reference
} // namespace tests
} // namespace simdutf
