#include "simdutf/common_defs.h"

namespace simdutf {
namespace tests {
namespace reference {

simdutf_warn_unused bool validate_utf32_to_latin1(const char32_t *buf,
                                                  size_t len) noexcept;

}
} // namespace tests
} // namespace simdutf
