#ifndef SIMDUTF_FALLBACK_BITMANIPULATION_H
#define SIMDUTF_FALLBACK_BITMANIPULATION_H

#include "simdutf.h"
#include <limits>

namespace simdutf {
namespace SIMDUTF_IMPLEMENTATION {
namespace {} // unnamed namespace
} // namespace SIMDUTF_IMPLEMENTATION
} // namespace simdutf

#endif // SIMDUTF_FALLBACK_BITMANIPULATION_H
